#include "BF_GSS.h"
#include <../config.h>
#include <stdio.h>
#include <iostream>

namespace BF
{
    class InvalidStateException : public std::exception
    {
    };

    GSS::GSS(int _numFreq, int _samplingRate, double _soundSpeed, bool _muAdp, double _mu, bool _mu_ssAdp, double _mu_ss, double _scale, NlFuncType _nlFuncType) :
	BF(_numFreq, _samplingRate, _soundSpeed),
	muAdp(_muAdp),
	mu(_mu),
	mu_ssAdp(_mu_ssAdp),
	mu_ss(_mu_ss),
	scale(_scale),
	nlFuncType(_nlFuncType)
    {
    }

    GSS::~GSS()
    {
    }

/*== source from adaptBF.m

  % For GSS, the original paper is 
  % [3] Parra, L. C., et al.: “Geometric source separation:
  % Merging convolutive source separation with geometric beamforming”,
  % IEEE Trans. SAP Vol.10, No.6, pp.352-362, 2002.
  % This program is based on instanteneous and generalized version of GSS, 
  % that is writeen in 
  % [4] H. Nakajima, et al.: "Blind Source Separation With Parameter-Free
  % Adaptive Step-Size Method for Robot Audition", 
  % IEEE Trans. ASL Vol.18, No.6, pp.1476-1485, 2010.
  % GICA and GHDSS are also included in [4].
      if isempty(G) % Initialization process
        G = zeros(NSrc, NMic, NFreq); 
        for nf = 1:NFreq
          for ns = 1:NSrc % for each source
            h = Adata.H(:, ns, nf); % Target source TF
            G(ns, :, nf) = pinv(h);
          end
        end
        return
      else % update process
        for nf = 1:NFreq % for each frequency bin
          W = G(:, :, nf);
    
          % fo GC-based calculation
          D = Adata.H(:, :, nf);
          myu_gc = Adata.myu;
          Egc = diag(diag(W*D - eye(NSrc)));
          Jpgc = Egc * D';
          
          % for SS-based calculation
          x = Adata.x(:, nf);
          y = Adata.y(:, nf);
          myu_ss = Adata.myu_ss;
          
          if method == 8 % For GSS (linear func.)
            Ryy = y*y';
          else % For ICA, HDSS (use nonlinear func.)
            switch Adata.nlfunc
            case 0 % tanh
              ya = abs(y);
              py = tanh(Adata.scal*ya) .* (y./ya);
            case 1 % x^3
              ya = abs(y);
              py = ya.^3 .* (y./ya);
            end
              Ryy = py * y';                          
          end
          Ess = Ryy - diag(diag(Ryy));
          
          if method == 8 % For GSS (linear func.)
            Jpss = 2 * Ess * W * x * x';
          elseif method == 9 % For ICA
            Jpss = Ess * W;
          elseif method == 10 % For GHDSS
            Jpss = 2 * Ess * py * x';
          end
          
          % update based on GSS, GICA, GHDSS
          W = W - myu_ss * Jpss - myu_gc * Jpgc;
          
          G(:, :, nf) = W; % Store separation matrix of updated.
          
        end
        return
      end % For if initialization
  
*/
    void
    GSS::updateSources(const std::vector<SourceLocation> &sourcePositions)
    {
	int numSrc = sourcePositions.size();

	/*== 直近フレームでの音源と、このフレームでの音源との対応をとる ==*/
	std::vector<int> lastSourceIndices; // i 番めの要素は、このフレームでの i 番目の音源（配列の i 番目の要素という意味）が、直近フレームでの何番目の音源（同）に相当するかを表す。
	for(int i = 0; i < numSrc; i++){
	    const SourceLocation &sourcePosition = sourcePositions[i];

	    /* 直近フレームの音源から、同じ ID を持つものを探す */
	    int j;
	    for(j = 0; j < this->sourcePositions.size(); j++){
		if(sourcePosition.sourceID == this->sourcePositions[j].sourceID){
		    break;
		}
	    }
	    
	    if(j < this->sourcePositions.size()){
		/* 見つかった */
		lastSourceIndices.push_back(j);
	    }else{
		/* 見つからなかった */
		lastSourceIndices.push_back(-1);
	    }
	}
	
	/*== 伝達関数を再計算する ==*/
	std::vector<Matrix> newHs, newGs;
	this->BF::getTransformFunctions(sourcePositions, newHs);

	for(int i = 0; i < this->numFreq; i++){
	    const Matrix &H = newHs[i];
	    int numMic = H.getRowCount();

	    Matrix G(numSrc, numMic);
	    for(int j = 0; j < numSrc; j++){
		/* 入力の j 番目の音源と同じIDの音源が、直近フレームに見当たらない？ */
		const int lastSourceIndex = lastSourceIndices[j];
		if(lastSourceIndex >= 0){
		    G.setRow(j, this->separationMatrices[i].getRow(lastSourceIndex));
		}else{
		    Matrix h = H.getColumn(j);
		    G.setRow(j, h.pseudoInverse());
		}
	    }

	    newGs.push_back(G);
	}

	/*== 結果の格納 ==*/
	this->sourcePositions = sourcePositions;
	newHs.swap(this->transformFunctions);
	newGs.swap(this->separationMatrices);
    }

    void
    GSS::adaptSeparationMatrices()
    {

	int numSrc = this->sourcePositions.size();

	std::vector<Matrix> newGs;
	for(int i = 0; i < this->numFreq; i++){
	    const Matrix &H = this->transformFunctions[i];
	    int numMic = H.getRowCount();

	    Matrix        W    = this->separationMatrices[i];
	    const Matrix &D    = H;
	    Matrix        Egc  = (W * D - Matrix::I(numSrc)).diagonalElements().diagonalMatrix();
	    Matrix        Jpgc = Egc * D.transposed();
	    Matrix        x    = this->inputSpectrum.getColumn(i);
	    Matrix        y    = this->outputSpectrum.getColumn(i);
	    
	    // Adaptive Step Size (GSS)
	    double mu;
	    if(this->muAdp){
		if(std::abs(Jpgc.elementwiseNorm().getElementSum()) < BF::BF_EPS){
		      mu = 
		      std::abs(Egc.elementwiseNorm().getElementSum()) / 2.0 /
		      (std::abs(Jpgc.elementwiseNorm().getElementSum()) + BF::BF_EPS) *
		      this->mu;

		}else{
		      mu = 
		      std::abs(Egc.elementwiseNorm().getElementSum()) / 2.0 /
		      std::abs(Jpgc.elementwiseNorm().getElementSum()) *
		      this->mu;

		}
	    }else{
	      mu = this->mu;
	    }

	    Matrix ryy;
	    Matrix py(y.getRowCount(), 1);
	    switch(this->getBfAlgorithmType()){
#ifdef GSS_ENABLE
	    case BfGSS:
		ryy = y * y.transposed();
		break;
	    case BfGHDSS:
#endif
	    case BfGICA:
		{
		    switch(this->nlFuncType){
		    case NlFuncTanh:
			{
			    for(int i = 0; i < y.getRowCount(); i++){
				Complex v = y.getElementAt(i, 0);
				double  a = std::abs(v);
				if(a < BF_EPS){
				    py.setElementAt(i, 0, 0);
				}else{
					if(a < BF::BF_EPS){ 
						
						py.setElementAt(i, 0, tanh(this->scale * a) * (v / (a + BF::BF_EPS)));
					}
					else{

						if (tanh(this->scale * a )  > 1.0 )
						{py.setElementAt(i, 0, 1.0 * (v / a));}
						else{
				    		py.setElementAt(i, 0, tanh(this->scale * a) * (v / a));}
					}
				}
//				fprintf(stdout, "a=%f\n", a);
			    }
//			    py.dump(stdout, "py");
			    break;
			}

		    case NlFuncCube:
			{
			    for(int i = 0; i < y.getRowCount(); i++){
				Complex v = y.getElementAt(i, 0);
				py.setElementAt(i, 0, v * v * std::conj(v));
			    }
			    break;
			}
		    }
		    ryy = py * y.transposed();
		}
		break;

	    default:
		throw InvalidStateException();
	    }

	    Matrix Ess = ryy - ryy.diagonalElements().diagonalMatrix();
	    Matrix Jpss;
	    switch(this->getBfAlgorithmType()){
#ifdef GSS_ENABLE
	    case BfGSS:   Jpss = 2 * Ess * W * x * x.transposed(); break;
	    case BfGHDSS: Jpss = 2 * Ess * py * x.transposed(); break;
#endif
	    case BfGICA:  Jpss = Ess * W; break;
	    default:
		throw InvalidStateException();
	    }

	    // Adaptive Step Size (DS, ICA, HDSS)
	    double mu_ss;
	    Matrix den1;
	    Matrix den2;
	    if(this->mu_ssAdp){
	      switch(this->getBfAlgorithmType()){
#ifdef GSS_ENABLE
	      case BfGSS: 
		 if(std::abs(Jpss.elementwiseNorm().getElementSum()) < BF::BF_EPS){
			  mu_ss = 
				  std::abs(Ess.elementwiseNorm().getElementSum()) / 2.0 /
				  (std::abs(Jpss.elementwiseNorm().getElementSum()) + BF::BF_EPS) *
				  this->mu_ss;
		  }
		  else
		  {
			mu_ss = 
				  std::abs(Ess.elementwiseNorm().getElementSum()) / 2.0 /
				  std::abs(Jpss.elementwiseNorm().getElementSum()) *
				  this->mu_ss;
		  }
		break;	      
	      case BfGHDSS: 
		 if(std::abs(Jpss.elementwiseNorm().getElementSum()) < BF::BF_EPS){
				mu_ss = 
				  std::abs(Ess.elementwiseNorm().getElementSum()) / 2.0 /
				  (std::abs(Jpss.elementwiseNorm().getElementSum()) + BF::BF_EPS) *
				  this->mu_ss;
		  }
		  else
		  {
				mu_ss = 
				  std::abs(Ess.elementwiseNorm().getElementSum()) / 2.0 /
				  std::abs(Jpss.elementwiseNorm().getElementSum()) *
				  this->mu_ss;
		  }
		break;
#endif
	      case BfGICA:
		den1 = Ess * W;
		den2 = 2 * Ess * py * x.transposed();
		if((den1.elementwiseFMultiplied(den2).getElementSum().real() ) < BF::BF_EPS)
		{
			mu_ss = 
			  std::abs(Ess.elementwiseNorm().getElementSum()) / 2.0 /
			  (den1.elementwiseFMultiplied(den2).getElementSum().real() + BF::BF_EPS) * 
			  this->mu_ss; 
		}
		else
		{
			mu_ss = 
			  std::abs(Ess.elementwiseNorm().getElementSum()) / 2.0 /
			  den1.elementwiseFMultiplied(den2).getElementSum().real() * 
			  this->mu_ss; 
		}
		break;
	      default:
		throw InvalidStateException();
	      }
	    }else{
	      mu_ss = this->mu_ss;
	    }

	    //ess.dump(stdout, "jpss");

	    /*
	    D.dump(stdout, "D");
	    W.dump(stdout, "W");
	    x.dump(stdout, "x");
	    y.dump(stdout, "y");
	    Ess.dump(stdout, "Ess");
	    Jpss.dump(stdout, "Jpss");
	    Jpgc.dump(stdout, "Jpgc");

	    (2 * Ess).dump(stdout, "2 * Ess");
	    (2 * Ess * W).dump(stdout, "2 * Ess * W");
	    (2 * Ess * W * x).dump(stdout, "2 * Ess * w * x");
	    (2 * Ess * W * x * x.transposed()).dump(stdout, "2 * Ess * w * x * x'");
	    */

#if ENABLE_JDEBUG
	    double Jc = 
	      std::abs(Ess.elementwiseNorm().getElementSum()) + 
	      std::abs(Egc.elementwiseNorm().getElementSum());
	    double Jc1 = 
	      std::abs(Ess.elementwiseNorm().getElementSum());
	    double Jc2 = 
	      std::abs(Egc.elementwiseNorm().getElementSum());
	    std::cout << mu_ss << " " << mu << " " << Jc << " ";		
#endif

	    newGs.push_back(W - mu_ss * Jpss - mu * Jpgc);
	}

#if ENABLE_JDEBUG
	std::cout << std::endl;
#endif
	//newSeparationMatrices[0].dump(stdout, "G[freq=0]");

	/*== 結果の格納 ==*/
	newGs.swap(this->separationMatrices);
    }
}
